/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Tempura routines for lists.  
 * The declaration list(L,n) specifies that L is a list of fixed (static)
 * size n, whereas List(L,n) specifies that L is a list of size <= n (as yet
 * unspecified). 
 */

/* Declare a variable to always hold a FIXED-LENGTH list. */
define fixed_list(L,n) = {
    list(L,n) and stable struct(L)
}.

/* Declare a variable to always hold a VARIABLE-LENGTH list. */
define flex_list(L,n) = {
    List(L,n) and stable struct(L)
}.

define head(L) = L[0].
define tail(L) = L[1..|L|].

/* exchange two elements of a list */
define swap_list(L,i,j) = {
    forall k < |L|: {
	if k=i then L[k]:=L[j]
	else if k=j then L[k]:=L[i]
	else L[k]:=L[k]
    }
}.

/* extend a list by one element */
define extend_list(L,d) = {
    list(next L, |L|+1) and
    forall i<|L|+1: if i<|L| then L[i]:=L[i] else L[i]:=d
}.

define part_step(L,key,I,J) = {
    if L[I] < key
    then {
	stable L
	and I:=I+1 and J:=J
    }
    else {
	swap_list(L,I,J-1)
	and I:=I and J:=J-1
    }
}.

/* partition a list into two parts */
define partition(L,key,left_len) = {
    exists I,J : {
	I = 0 and J=|L|
	and while I<J do {
	    skip and
	    part_step(L,key,I,J)
	};
	{
	    empty and left_len=I
	}
    }
}.

/* test routine for partitioning */
define part_test(part_list) = {
    exists L,key,left_len,State_Num: {
	fixed_list(L,|part_list|)
	and L=part_list and key=part_list[0]
	and partition(L,key,left_len)
	and State_Num = 0 and State_Num gets State_Num+1
	and box format("State%3D:  List=%t\n",State_Num,L)
	and fin format("Number of list elements < first element = %d\n",
	    left_len)
    }
}.

/* partitioning used by quicksort routines */
define quick_partition(L,pivot) = {
    {
	partition(L[0 to |L|-1],L[|L|-1],pivot)
	and stable L[|L|-1]
    };
    {
	skip and swap_list(L,pivot,|L|-1)
    }
}.

