/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: tempura.h,v 2.16 2011/10/03 14:14:30 acau Exp $
 */

/********************\
*                    *
*  Type definitions  *
*                    *
\********************/

struct node;
struct env_val;
struct bnd_val;
struct loc_val;
struct mem_val;
struct block_val;
struct list_val;
struct name_val;
struct link_val;
struct symbol_val;

typedef struct node NODE, *NODEPTR;

typedef struct env_val ENVVAL;
typedef struct bnd_val BNDVAL;
typedef struct loc_val LOCVAL;
typedef struct mem_val MEMVAL;
typedef struct block_val BLKVAL;
typedef struct list_val LSTVAL;
typedef struct name_val NMVAL;
typedef struct link_val LNKVAL;
typedef struct symbol_val SYMVAL;

struct env_val {
    unsigned env_level;
    unsigned int env_size;
    unsigned int env_history;
    NODEPTR id_list;
};

struct bnd_val{
    NODEPTR env;
    NODEPTR blk; 
};

struct loc_val {
    NODEPTR ptr;
    unsigned to_next;
};

struct mem_val {
    unsigned int mem_no ;
    unsigned is_static : 1;
    unsigned is_indirect : 1;
    int time;
    unsigned is_undefined : 1;
    unsigned has_default : 1;
    unsigned int history;
};

struct block_val {
    unsigned int blk_size;
    NODEPTR blk_base;
};

struct list_val {
    NODEPTR blk;
    NODEPTR size;
    NODEPTR base;
};

struct name_val {
    char *name;
    int level;
    int off;
};

struct link_val {
    struct node *c1;
    struct node *c2;
    struct node *c3;
    struct node *c4;
};

struct symbol_val {
    char *name;
};

typedef union node_val {
    LNKVAL lnkval;
    long ival;
    char *str;
    double flval;
    int bval;
    ENVVAL envval;
    BNDVAL bndval;
    MEMVAL memval;
    BLKVAL blkval;
    LSTVAL lstval;
    NMVAL  nmval;
    SYMVAL symval;
    LOCVAL locval;
} NODEVAL;

struct node {
    unsigned mark : 1;
    unsigned char type;
    struct node *next_node;
    NODEVAL nval;
} ;
#define nodep(X) ((NODEPTR) (X))


typedef void (*PFI)();

typedef struct heap_block {
    struct heap_block *next_block;
    struct node *base;
    struct node *top;
} HEAPNODE, *HEAP;
#define heapp(X) ((HEAP) (X))

typedef struct es_item {
    NODEPTR env_base;
    NODEPTR env;
} ENVSTACK, *ENVSTACKPTR;

typedef struct reduce_stack {
    NODEPTR n;
} REDUCESTACK, *REDUCESTACKPTR;

#define node_mark(E) (E) -> mark
#define node_type(E) (E) -> type
#define node_next(E) (E) -> next_node
#define node_pnext(E)  (E).next_node

#define node_lnk_1(E) (E) -> nval.lnkval.c1
#define node_lnk_2(E) (E) -> nval.lnkval.c2
#define node_lnk_3(E) (E) -> nval.lnkval.c3
#define node_lnk_4(E) (E) -> nval.lnkval.c4

#define node_ival(E) (E) -> nval.ival

#define node_strval(E) (E) -> nval.str

#define node_flval(E) (E) -> nval.flval

#define node_bval(E) (E) -> nval.bval

#define node_env_level(E) (E) -> nval.envval.env_level
#define node_env_size(E) (E) -> nval.envval.env_size
#define node_env_idlist(E) (E) -> nval.envval.id_list
#define node_env_history(E) (E) -> nval.envval.env_history

#define node_bnd_env(E) (E) -> nval.bndval.env
#define node_bnd_blk(E) (E) -> nval.bndval.blk

#define node_mem_memno(E) (E) -> nval.memval.mem_no
#define node_mem_isstatic(E) (E) -> nval.memval.is_static
#define node_mem_isindirect(E) (E) -> nval.memval.is_indirect
#define node_mem_time(E) (E) -> nval.memval.time
#define node_mem_isundefined(E) (E) -> nval.memval.is_undefined
#define node_mem_hasdefault(E) (E) -> nval.memval.has_default
#define node_mem_history(E) (E) -> nval.memval.history

#define node_blk_size(E) (E) -> nval.blkval.blk_size
#define node_blk_base(E) (E) -> nval.blkval.blk_base

#define node_lst_blk(E) (E) -> nval.lstval.blk
#define node_lst_size(E) (E) -> nval.lstval.size
#define node_lst_base(E) (E) -> nval.lstval.base

#define node_nm_name(E) (E) -> nval.nmval.name
#define node_nm_level(E) (E) -> nval.nmval.level
#define node_nm_off(E) (E) -> nval.nmval.off

#define node_sym_name(E) (E) -> nval.symval.name

#define node_loc_ptr(E) (E) -> nval.locval.ptr
#define node_loc_tonext(E) (E) -> nval.locval.to_next


#define save_state(BP) \
        int old_env_base = EnvBase;\
        int old_env_top = EnvTop;\
        REDUCESTACKPTR old_reduce_stack_top = ReduceStackTop;\
        int old_reduce_depth = ReduceDepth;\
        NODEPTR old_empty_loc = EmptyLoc;\
        unsigned old_next_flag = NextFlag;\
        unsigned old_prev_flag = PrevFlag;\
        unsigned old_assign_ahead_flag = AssignAheadFlag;\
        unsigned old_break_is_abort_flag = BreakIsAbortFlag;\
        unsigned old_run_flag = RunFlag;\
        unsigned old_is_initial = IsInitial;\
        jmp_buf break_point;\
        BP = &break_point;\
        setjmp(break_point);\
        EnvBase = old_env_base;\
        EnvTop = old_env_top;\
        ReduceStackTop = old_reduce_stack_top;\
        ReduceDepth = old_reduce_depth;\
        EmptyLoc = old_empty_loc;\
        NextFlag = old_next_flag;\
        PrevFlag = old_prev_flag;\
        AssignAheadFlag = old_assign_ahead_flag;\
        BreakIsAbortFlag = old_break_is_abort_flag;\
        RunFlag = old_run_flag;\
        IsInitial = old_is_initial;              \
        

#define restore_state(BP,val) longjmp(*(BP),val)

#define F 0
#define T 1
#define UNDEF -1

#define VAL 1
#define REF 2
#define STMT 3

#define syntax_error0(S) syntax_error(S,NULL)
#define syntax_error1(S,X) syntax_error(S,X)

#define error_break1(S,X) error_break2(S,X,NULL)
#define error_break1d(S,X) error_break2e(S,X)
#define error_break0(S) error_break2(S,NULL,NULL)

#define warn0(S) warn1(S,NULL)

#define assert0(C,S) if(!(C)) error_break0(S)
#define assert1(C,S,X) if(!(C)) error_break1(S,X)
#define assert2(C,S,X,Y) if(!(C)) error_break2(S,X,Y)
#define assert2d(C,S,X,Y) if(!(C)) error_break2d(S,X,Y)

#define result(E) ReduceStackTop->n=(E)
#define ref_result(E) ReduceStackTop->n=(E) 



#define node0(U) node(U,NULL,NULL,NULL,NULL)
#define node1(U,V) node(U,V,NULL,NULL,NULL)
#define node2(U,V,W) node(U,V,W,NULL,NULL)
#define node3(U,V,W,X) node(U,V,W,X,NULL)
#define node4(U,V,W,X,Y) node(U,V,W,X,Y)

#define conj(U,V) node3(iAND,False,U,node3(iAND,False,V,NULL))
#define conjtf(U,V) node3(iAND,True,U,node3(iAND,False,V,NULL))
#define assign(L,V) node2(EQ,L,copy_val(V))

#define max(A,B) (((A) > (B)) ? (A) : (B))
#define min(A,B) (((A) < (B)) ? (A) : (B))

#define mod(A,B) (A<0?(A+((B-A)/B)*B)%B:A%B)

#define MEMCELLSIZE 5
#define cycle2(E) ((E == 1) ? 0 : 1)
#define cycle(E) ((E == 2) ? 3 : (E==3)? 4 : 2)
#define switch_memory CurrBit = cycle2(CurrBit);NextBit = cycle2(NextBit);   
/*#define switch_memory CurrBit = addm2(CurrBit);NextBit = addm2(NextBit); \
  PrevBit = addm2(PrevBit)*/
/*#define switch_memory CurrBit = 1-CurrBit;NextBit = 1-NextBit*/
/*#define zero_loc(X) ((X)-node_mem_memno(X))*/
/*#define curr_loc0(X) ((X)+CurrBit)*/
/*
#define curr_loc(X) (((X)-node_mem_memno(X))+CurrBit)
#define next_loc(X) (((X)-node_mem_memno(X))+NextBit)
#define prev_loc(X) (((X)-node_mem_memno(X))+PrevBit)
*/
/*#define prevcount(X,Y) 2 + ((X+Y) % 3)*/
#define prevcount1(X,Y,Z) 2 + ((X+Z) % (2+Y))
#define curr_loc(X) ((X)+(int)(CurrBit-(node_mem_memno(X))))
#define next_loc(X) ((X)+(int)(NextBit-(node_mem_memno(X))))
/*#define prev_loc(X) ((X)+(int)(prevcount(Clock,0)-(node_mem_memno(X))))
#define prevnext_loc(X) ((X)+(int)(prevcount(Clock,1)-(node_mem_memno(X))))
#define prevprev_loc(X) ((X)+(int)(prevcount(Clock,2)-(node_mem_memno(X))))*/
#define prev_loc1(X,Z) ((X)+(int)(prevcount1(Clock,node_mem_history(X),Z)-(node_mem_memno(X))))
/*
#define is_curr(X) (node_mem_memno(X) == CurrBit)
#define is_next(X) (node_mem_memno(X) == NextBit)
#define is_prev(X) (node_mem_memno(X) == PrevBit)*/
/*#define is_prev(X) (node_mem_memno(X) == prevcount(Clock,0))
#define is_prevnext(X) (node_mem_memno(X) == prevcount(Clock,1))
#define is_prevprev(X) (node_mem_memno(X) == prevcount(Clock,2))*/
#define is_prev1(X,Z) (node_mem_memno(X) == prevcount1(Clock,node_mem_history(X),Z))
#define is_curr(X) (node_mem_memno(X) == CurrBit)
#define is_next(X) (node_mem_memno(X) == NextBit)
/*#define adj_loc(X) ((NextFlag)? ((PrevFlag)? curr_loc(X): next_loc(X)):((PrevFlag)?prevprev_loc(X):curr_loc(X)))*/
#define adj_loc1(X) \
   ((NextFlag)?\
    ((PrevFlag>0 && PrevFlag-1<=node_mem_history(X))?     \
         ((PrevFlag==1)?\
            curr_loc(X):\
          prev_loc1(X,3+node_mem_history(X)-PrevFlag) \
         ) :\
       ((PrevFlag==0)? next_loc(X): NULL)\
       ):\
    ((PrevFlag>0 && PrevFlag<=node_mem_history(X))?     \
     prev_loc1(X,2+node_mem_history(X)-PrevFlag):   \
     ((PrevFlag==0)? curr_loc(X): NULL)\
    )\
   )

/*#define adj_loc2(X,Y,Z) ((Y)? ((Z)? curr_loc(X): next_loc(X)):((Z)?prevprev_loc(X):curr_loc(X)))*/
/*#define offset(X) ((X)*MEMCELLSIZE)*/
#define offset2(X,Y) ((X)*(4+Y))
#define cell_no(X) (X)
/*#define id_list_size(X) (((X)==NULL)?0:(X)->offset+1)*/
#define mem_loc(X) (EnvStack[EnvBase+node_nm_level(X)].env_base+node_nm_off(X))
#define history(X) (node_env_history(EnvStack[EnvBase+node_nm_level(X)].env))
#define is_indirect_mem(X) (node_mem_isindirect(X))
/*#define dereference(L) while (is_indirect_mem (L)) (L) = memoryp((L)->val)*/
#define dereference(L) while (is_indirect_mem (L)) (L) = node_next(L)
#define local_mem(A) mem_block1(A,1,1)
#define local_mem1(A,H) mem_block1(A,1,H)
#define GlobalMemUsed node_env_size(EnvStack[EnvBase].env)

#define get_int(X) (node_ival(get_mem(X)))
#define get_list(X) get_mem(X)
#define put_list(L,X) put_mem1(L,X)

#define list_size(L) node_lst_size(L)
#define list_base(L) node_lst_base(L)
#define list_blk(L)  node_lst_blk(L)

#define is_identifier(X) ((node_type(X) == STATE)||(node_type(X) == STATIC))
#define is_boolean(X) (node_type(X) == BOOL)
#define is_integer(X) (node_type(X) == NAT)
#define is_string(X) (node_type(X) == STRING)
#define is_float(X) (node_type(X) == FLOAT)
#define is_lambda(X) (node_type(X) == LAMBDA)
#define is_list(X) (node_type(X) == LIST)

#define curr_val(X) is_list(X) ? curr_list(X) : (X)
#define next_val(X) is_list(X) ? next_list(X) : (X)

/*#define prev_val(X) is_list(X) ? prev_list(X) : (X)
#define prevnext_val(X) is_list(X) ? prevnext_list(X) : (X)
#define prevprev_val(X) is_list(X) ? prevprev_list(X) : (X)
*/
/*#define adj_list(X)  ((NextFlag)? ((PrevFlag)? curr_list(X): next_list(X)):((PrevFlag)?prevprev_list(X):curr_list(X)))*/

#define pass_verbosity (Verbosity > 3)
#define state_verbosity (Verbosity > 2)
#define store_verbosity (Verbosity > 1)
#define exec_verbosity (Verbosity > 0)
#define state_debug DebugLevel == 2
#define reduce_debug DebugLevel > 2

#define output_char(C) {putchar(C);CharsOutput++;}
#define space output_char(' ')
#define fspace foutput_char(' ')
#define newline output_char('\n')
#define fnewline foutput_char('\n')
#define output_state_number if (RunFlag && StatePrintFlag) { \
                                printf("State %3d: ",StateNumber); \
                                CharsOutput += 11;}

#define push_empty(X) NODEPTR old_empty_loc=EmptyLoc;EmptyLoc=curr_loc(X)
#define pop_empty EmptyLoc=old_empty_loc;
    

#define push_expn(E)                            \
    if (++ReduceDepth > MAXREDUCESTACKSIZE) {\
        ReduceStackTop -> n = E;\
        error_break0 ("reduction stack full");\
    }\
    else {\
        (++ReduceStackTop) -> n = E;\
        MaxReduceDepth = max (MaxReduceDepth, ReduceDepth);\
    }\
    PassReduceCount += 1

#define pop_expn (ReduceDepth--,(ReduceStackTop--) -> n)
#define pop_ref (ReduceDepth--,(ReduceStackTop--) -> n)

#define always(EXP) \
{\
    NODEPTR e1 = EXP;\
    if (e1 != True) {\
        if (Ready)\
            result (conj (e1, e));\
        else\
            result (conjtf (e, e1));\
    }\
}

#define keep(EXP) \
{\
    if ((Ready = empty_defined ())) {\
        if (get_empty () == True)\
            result (True);\
        else {\
            NODEPTR e1 = EXP;\
            if (e1 != True) {\
                if (Ready)\
                    result (conj (e1, e));\
	        else\
                    result (conjtf (e, e1));\
	    }\
        }\
    }\
}

#define push_assign_ahead unsigned old_assign_ahead_flag = AssignAheadFlag;\
        AssignAheadFlag = F;
#define pop_assign_ahead AssignAheadFlag = old_assign_ahead_flag;
